#region File Header
//-----------------------------------------------------------------------------
// SibLib
//
// Copyright (C) 2010 Julien Villers
// This program is distributed under the terms of the 
// GNU Lesser General Public License (LGPL).
// See Docs/lgpl.txt and Docs/gpl.txt
//
// This file is based on a sample from the book 'XNA Unleashed' (C) Chad Carter,
// with numerous changes by Julien Villers.
//-----------------------------------------------------------------------------
#endregion

using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;

namespace SibLib.Sound
{
    /// <summary>
    /// Helper interface for Game service support.
    /// </summary>
    public interface ISoundManager
    {
        #region Properties
        AudioEngine AudioEngine
        {
            get;
        }

        AudioListener AudioListener
        {
            get;
        }

        bool RepeatPlayList
        {
            get;
            set;
        }
        #endregion

        void SetGlobalVariable(string name, float amount);
        void SetVolume(string categoryName, float volumeAmount);
        float GetVolume(string categoryName);

        void PauseCategory(string categoryName);
        void ResumeCategory(string categoryName);

        bool IsPlaying(string cueName);
        void Play(string cueName);
        void Pause(string cueName);
        void Resume(string cueName);
        void Toggle(string cueName);
        void StopAll();
        void Stop(string cueName);

        void StartPlayList(string[] playList);
        void SetPlayList(string[] playList);
        void StartPlayList(string[] playList, int startIndex);
        void StartPlayList(int startIndex);
        void StartPlayList();
        void StopPlayList();

        Cue GetCue(string cueName);
        void PlayCue(string cueName, AudioEmitter emitter);
        void PlayCue(string cueName, Vector3 position);
    }

    // TODO: notification on music change, to be able to display its name.

    /// <summary>
    /// This is a tool to manage everything related to sound:
    ///   - Sound FX
    ///   - Music
    ///   - Volume/Disabling
    /// </summary>
    /// <remarks>
    /// Don't try to play anything in the first frame of the game.          // TODO: this sounds more a bug than a feature, doesn't it?
    /// </remarks>
    public class SoundManager : GameComponent, ISoundManager
    {
        #region Private Members
        private AudioEngine m_Engine;
        private WaveBank m_WaveBankSFX;
        private SoundBank m_SoundBankSFX;
        private WaveBank m_WaveBankMusic;
        private SoundBank m_SoundBankMusic;

        private Dictionary<string, Cue> m_Cues = new Dictionary<string, Cue>();
        private Dictionary<string, AudioCategory> m_Categories = new Dictionary<string, AudioCategory>();
        private Dictionary<string, float> m_Volumes = new Dictionary<string, float>();

        private string[] m_PlayList;
        private int m_SongCurrent;
        private float m_SilenceCurrent;
        private Cue m_CueCurrent;
        private AudioEmitter m_AudioEmitter;        // Used by fire and forget 3D positioned sound.
        #endregion

        #region Properties
        public AudioEngine AudioEngine
        {
            get { return m_Engine; }
        }

        public AudioListener AudioListener
        {
            get;
            private set;
        }

        public bool RepeatPlayList
        {
            get;
            set;
        }

        public string ContentPath
        {
            get;
            set;
        }
        public string BankSFXName
        {
            get;
            set;
        }
        public string BankMusicName
        {
            get;
            set;
        }
        #endregion

        #region Constructors

        /// <summary>
        /// Fully parametrized constructor.
        /// Setup this as a service and a component of game.
        /// </summary>
        /// <param name="game">a Game instance</param>
        /// <param name="xactProjectName">the name of the XACT project</param>      // TODO: USELESS, should rather expose banks!
        /// <param name="xactFileName">the name of the XACT file, should be without .xap (safety included)</param>
        /// <param name="contentPath"></param>
        public SoundManager(Game game, string xactProjectName, string xactFileName, string contentPath)
            : base(game)
        {
            game.Services.AddService(typeof(ISoundManager), this);
            game.Components.Add(this);

            xactFileName = xactFileName.Replace(".xap", "");

            m_Engine = new AudioEngine(contentPath + xactFileName + ".xgs");
            ContentPath = contentPath;
            BankSFXName = "Game";
            BankMusicName = "Music";

            AudioListener = new AudioListener();
            m_AudioEmitter = new AudioEmitter();

            RepeatPlayList = true;
        }

        public SoundManager(Game game, string xactProjectName)
            : this(game, xactProjectName, xactProjectName)
        {
        }

        public SoundManager(Game game, string xactProjectName, string xactFileName)
            : this(game, xactProjectName, xactFileName, game.Content.RootDirectory + @"\Sound\")
        { 
        }

        #endregion

        public override void Initialize()
        {
            m_WaveBankSFX = new WaveBank(m_Engine, ContentPath + "WaveBank" + BankSFXName+".xwb");
            m_SoundBankSFX = new SoundBank(m_Engine, ContentPath + "SoundBank" + BankSFXName + ".xsb");
            m_WaveBankMusic = new WaveBank(m_Engine, ContentPath + "WaveBank" + BankMusicName + ".xwb", 0, 32);
            m_SoundBankMusic = new SoundBank(m_Engine, ContentPath + "SoundBank" + BankMusicName + ".xsb");
            
            base.Initialize();
        }

        protected override void Dispose(bool disposing)
        {
            m_SoundBankSFX.Dispose();
            m_WaveBankSFX.Dispose();
            m_WaveBankMusic.Dispose();
            m_SoundBankMusic.Dispose();
            m_Engine.Dispose();

            m_PlayList = null;
            m_CueCurrent = null;
            m_Cues = null;
            m_SoundBankSFX = null;
            m_WaveBankSFX = null;
            m_Engine = null;
            AudioListener = null;
            m_AudioEmitter = null;

            base.Dispose(disposing);
        }

        #region Update

        public override void Update(GameTime gameTime)
        {
            m_Engine.Update();

            // Playlist update
            if (m_SilenceCurrent > 0.0f)
            {
                m_SilenceCurrent -= (gameTime.ElapsedGameTime.Milliseconds / 1000.0f);

                if (m_SilenceCurrent <= 0.0f)
                {
                    m_SongCurrent++;
                    PlaylistPlay();
                }
            }
            else if (m_CueCurrent != null) 
            {
                //check current cue to see if it is playing
                //if not, go to next cue in list
                if (!m_CueCurrent.IsPreparing && !m_CueCurrent.IsPlaying)
                {
                    m_SongCurrent++;

                    PlaylistPlay();
                }
            }

            base.Update(gameTime);
        }

        #endregion

        public void SetGlobalVariable(string name, float amount)
        {
            m_Engine.SetGlobalVariable(name, amount);
        }

        private void CheckCategory(string categoryName)
        {
            if (!m_Categories.ContainsKey(categoryName))
            {
                m_Categories.Add(categoryName, m_Engine.GetCategory(categoryName));
                m_Volumes.Add(categoryName, 1.0f);
            }
        }

        public void SetVolume(string categoryName, float volumeAmount)
        {
            CheckCategory(categoryName);

            m_Categories[categoryName].SetVolume(volumeAmount);
            m_Volumes[categoryName] = volumeAmount;
        }

        public float GetVolume(string categoryName)
        {
            CheckCategory(categoryName);

            return m_Volumes[categoryName];
        }

        public void PauseCategory(string categoryName)
        {
            CheckCategory(categoryName);

            m_Categories[categoryName].Pause();
        }

        public void ResumeCategory(string categoryName)
        {
            CheckCategory(categoryName);

            m_Categories[categoryName].Resume();
        }

        public bool IsPlaying(string cueName)
        {
            if (m_Cues.ContainsKey(cueName))
                return (m_Cues[cueName].IsPlaying);

            return (false);
        }

        public void Play(string cueName)
        {
            Cue prevCue = null;

            if (!m_Cues.ContainsKey(cueName))
                m_Cues.Add(cueName, m_SoundBankSFX.GetCue(cueName));
            else
            {
                //store our cue if we were playing
                if (m_Cues[cueName].IsPlaying)
                    prevCue = m_Cues[cueName];

                m_Cues[cueName] = m_SoundBankSFX.GetCue(cueName);
            }

            //if we werent playing, set previous to our current cue name
            if (prevCue == null)
                prevCue = m_Cues[cueName];

            try
            {
                m_Cues[cueName].Play();
            }
            catch (InstancePlayLimitException)
            {
                //hit limit exception, set our cue to the previous
                //and lets stop it and then start it up again ...
                m_Cues[cueName] = prevCue;

                if (m_Cues[cueName].IsPlaying)
                    m_Cues[cueName].Stop(AudioStopOptions.AsAuthored);

                Toggle(cueName);
            }
        }

        public void Pause(string cueName)
        {
            if (m_Cues.ContainsKey(cueName))
                m_Cues[cueName].Pause();
        }

        public void Resume(string cueName)
        {
            if (m_Cues.ContainsKey(cueName))
                m_Cues[cueName].Resume();
        }

        public void Toggle(string cueName)
        {
            if (m_Cues.ContainsKey(cueName))
            {
                Cue cue = m_Cues[cueName];

                if (cue.IsPaused)
                {
                    cue.Resume();
                }
                else if (cue.IsPlaying)
                {
                    cue.Pause();
                }
                else //played but stopped 
                {
                    //need to reget cue if stopped
                    Play(cueName);
                }
            }
            else //never played, need to reget cue
                Play(cueName);
        }

        public void StopAll()
        {
            foreach (Cue cue in m_Cues.Values)
                cue.Stop(AudioStopOptions.Immediate);
        }

        public void Stop(string cueName)
        {
            if (m_Cues.ContainsKey(cueName))
                m_Cues[cueName].Stop(AudioStopOptions.Immediate);
            m_Cues.Remove(cueName);
        }

        public void StartPlayList(string[] playList)
        {
            StartPlayList(playList, 0);
        }

        public void SetPlayList(string[] playList)
        {
            if (playList.Length == 0)
                return;

            this.m_PlayList = playList;
        }

        public void StartPlayList(string[] playList, int startIndex)
        {
            SetPlayList(playList);

            if (startIndex > playList.Length)
                startIndex = 0;

            StartPlayList(startIndex);
        }

        public void StartPlayList(int startIndex)
        {
            if (m_PlayList.Length == 0)
                return;

            m_SongCurrent = startIndex;
            PlaylistPlay();
        }

        public void StartPlayList()
        {
            StartPlayList(0);
        }

        public void StopPlayList()
        {
            if (m_CueCurrent != null)
            {
                m_CueCurrent.Stop(AudioStopOptions.Immediate);
                m_CueCurrent = null;
                m_SilenceCurrent = 0.0f;
            }
        }

        // JV: Additional methods

        private void PlaylistPlay()
        {
            if (m_SongCurrent >= m_PlayList.Length)
            {
                if (RepeatPlayList)
                {
                    m_SongCurrent = 0;
                }
                else
                {
                    StopPlayList();
                    return;
                }
            }

            if (m_PlayList[m_SongCurrent].StartsWith("SILENCE"))
            {
                m_SilenceCurrent = int.Parse(m_PlayList[m_SongCurrent].Substring(7)) + 0.1f;
            }
            else
            {
                m_SilenceCurrent = 0.0f;
                m_CueCurrent = m_SoundBankMusic.GetCue(m_PlayList[m_SongCurrent]);
                m_CueCurrent.Play();
            }
        }

        /// <summary>
        /// Obtain a new cue instance.
        /// Use this only when you need to keep a cue instance handy, 
        /// for instance for 3D positioning or for looping sounds that must be stopped manually.
        /// </summary>
        /// <param name="cueName">The name of the cue.</param>
        /// <returns>A new cue instance, if name was found.</returns>
        public Cue GetCue(string cueName)
        {
            return m_SoundBankSFX.GetCue(cueName);
        }

        /// <summary>
        /// Fire and forget 3D playback.
        /// </summary>
        /// <param name="cueName">Name of the cue</param>
        /// <param name="emitter">A setup AudioEmitter</param>
        public void PlayCue(string cueName, AudioEmitter emitter)
        {
            m_SoundBankSFX.PlayCue(cueName, AudioListener, emitter);
        }

        /// <summary>
        /// Fire and forget 3D playback.
        /// </summary>
        /// <param name="cueName">Name of the cue</param>
        /// <param name="position">The 3D position of the sound</param>
        public void PlayCue(string cueName, Vector3 position)
        {
            m_AudioEmitter.Position = position;
            m_SoundBankSFX.PlayCue(cueName, AudioListener, m_AudioEmitter);
        }
    }
}